# Copyright 2021 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul.lib import strings

from tests.base import BaseTestCase


class TestStrings(BaseTestCase):

    def test_unique_project_name(self):
        self.assertEqual(('project', 'project'),
                         strings.unique_project_name('project'))
        self.assertEqual(('project', 'project%2Fsubproject'),
                         strings.unique_project_name('project/subproject'))
        self.assertEqual(('project', 'project%2Fsub%2Fproject'),
                         strings.unique_project_name('project/sub/project'))
