// Copyright 2021 Red Hat, Inc
//
// Licensed under the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License. You may obtain
// a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

import * as React from 'react'
import PropTypes from 'prop-types'
import { Link } from 'react-router-dom'


function AutoholdNodes(props) {
  const {build, link } = props

  return (
    <span >
      <Link to={link}>{build}</Link>
    </span>
  )
}

AutoholdNodes.propTypes = {
  build: PropTypes.string,
  link: PropTypes.string,
}

export { AutoholdNodes }
